@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2024 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,6 +40,27 @@ static const sasl_callback_t client_callbacks[] = {
 
 static void upstream_unlink( LloadConnection *upstream );
 
+int
+lload_upstream_entry_cmp( const void *l, const void *r )
+{
+    return SLAP_PTRCMP( l, r );
+}
+
+static void
+linked_upstream_lost( LloadConnection *client )
+{
+    int gentle = 1;
+
+    CONNECTION_LOCK(client);
+    assert( client->c_restricted >= LLOAD_OP_RESTRICTED_UPSTREAM );
+    assert( client->c_linked_upstream );
+
+    client->c_restricted = LLOAD_OP_NOT_RESTRICTED;
+    client->c_linked_upstream = NULL;
+    CONNECTION_UNLOCK(client);
+    lload_connection_close( client, &gentle );
+}
+
 int
 forward_response( LloadConnection *client, LloadOperation *op, BerElement *ber )
 {
@@ -108,7 +129,7 @@ forward_final_response(
 
     op->o_res = LLOAD_OP_COMPLETED;
     if ( !op->o_pin_id ) {
-        OPERATION_UNLINK(op);
+        operation_unlink( op );
     }
 
     return rc;
@@ -118,22 +139,17 @@ static int
 handle_unsolicited( LloadConnection *c, BerElement *ber )
 {
     CONNECTION_ASSERT_LOCKED(c);
-
-    assert( c->c_state != LLOAD_C_INVALID );
-    if ( c->c_state == LLOAD_C_DYING ) {
-        CONNECTION_UNLOCK(c);
-        goto out;
+    if ( c->c_state != LLOAD_C_PREPARING ) {
+        c->c_state = LLOAD_C_CLOSING;
     }
-    c->c_state = LLOAD_C_CLOSING;
 
     Debug( LDAP_DEBUG_STATS, "handle_unsolicited: "
             "teardown for upstream connection connid=%lu\n",
             c->c_connid );
 
     CONNECTION_DESTROY(c);
-
-out:
     ber_free( ber, 1 );
+
     return -1;
 }
 
@@ -233,7 +249,21 @@ handle_one_response( LloadConnection *c )
         }
     }
     if ( op ) {
-        op->o_last_response = slap_get_time();
+        struct timeval tv, tvdiff;
+        uintptr_t diff;
+
+        gettimeofday( &tv, NULL );
+        if ( !timerisset( &op->o_last_response ) ) {
+            LloadBackend *b = c->c_backend;
+
+            timersub( &tv, &op->o_start, &tvdiff );
+            diff = 1000000 * tvdiff.tv_sec + tvdiff.tv_usec;
+
+            __atomic_add_fetch( &b->b_operation_count, 1, __ATOMIC_RELAXED );
+            __atomic_add_fetch( &b->b_operation_time, diff, __ATOMIC_RELAXED );
+        }
+        op->o_last_response = tv;
+
         Debug( LDAP_DEBUG_STATS2, "handle_one_response: "
                 "upstream connid=%lu, processing response for "
                 "client connid=%lu, msgid=%d\n",
@@ -933,6 +963,23 @@ upstream_init( ber_socket_t s, LloadBackend *b )
     /* We only add the write event when we have data pending */
     c->c_write_event = event;
 
+#ifdef BALANCER_MODULE
+    if ( b->b_monitor ) {
+        acquire_ref( &c->c_refcnt );
+        CONNECTION_UNLOCK(c);
+        checked_unlock( &b->b_mutex );
+        if ( lload_monitor_conn_entry_create( c, b->b_monitor ) ) {
+            RELEASE_REF( c, c_refcnt, c->c_destroy );
+            checked_lock( &b->b_mutex );
+            CONNECTION_LOCK(c);
+            goto fail;
+        }
+        checked_lock( &b->b_mutex );
+        CONNECTION_LOCK(c);
+        RELEASE_REF( c, c_refcnt, c->c_destroy );
+    }
+#endif /* BALANCER_MODULE */
+
     c->c_destroy = upstream_destroy;
     c->c_unlink = upstream_unlink;
 
@@ -979,6 +1026,14 @@ upstream_init( ber_socket_t s, LloadBackend *b )
     return c;
 
 fail:
+    if ( !IS_ALIVE( c, c_live ) ) {
+        /*
+         * Released while we were unlocked, it's scheduled for destruction
+         * already
+         */
+        return NULL;
+    }
+
     if ( c->c_write_event ) {
         event_del( c->c_write_event );
         event_free( c->c_write_event );
@@ -1001,7 +1056,7 @@ upstream_unlink( LloadConnection *c )
 {
     LloadBackend *b = c->c_backend;
     struct event *read_event, *write_event;
-    TAvlnode *root;
+    TAvlnode *root, *linked_root;
     long freed, executing;
 
     Debug( LDAP_DEBUG_CONNS, "upstream_unlink: "
@@ -1022,11 +1077,16 @@ upstream_unlink( LloadConnection *c )
     executing = c->c_n_ops_executing;
     c->c_n_ops_executing = 0;
 
+    linked_root = c->c_linked;
+    c->c_linked = NULL;
+
     CONNECTION_UNLOCK(c);
 
     freed = ldap_tavl_free( root, (AVL_FREE)operation_lost_upstream );
     assert( freed == executing );
 
+    ldap_tavl_free( linked_root, (AVL_FREE)linked_upstream_lost );
+
     /*
      * Avoid a deadlock:
      * event_del will block if the event is currently executing its callback,
@@ -1087,6 +1147,17 @@ upstream_destroy( LloadConnection *c )
 
     CONNECTION_LOCK(c);
     assert( c->c_state == LLOAD_C_DYING );
+
+#ifdef BALANCER_MODULE
+    /*
+     * Can't do this in upstream_unlink as that could be run from cn=monitor
+     * modify callback.
+     */
+    if ( !BER_BVISNULL( &c->c_monitor_dn ) ) {
+        lload_monitor_conn_unlink( c );
+    }
+#endif /* BALANCER_MODULE */
+
     c->c_state = LLOAD_C_INVALID;
 
     assert( c->c_ops == NULL );
