@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2018 The OpenLDAP Foundation.
+ * Copyright 2001-2022 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -120,7 +120,6 @@ monitor_send_children(
 		rc = test_filter( op, e, op->oq_search.rs_filter );
 		if ( rc == LDAP_COMPARE_TRUE ) {
 			rs->sr_entry = e;
-			rs->sr_flags = REP_ENTRY_MUSTRELEASE;
 			rc = send_search_entry( op, rs );
 			if ( rc ) {
 				for ( e = sub_ch; e != NULL; e = sub_nv ) {
@@ -129,9 +128,11 @@ monitor_send_children(
 					monitor_cache_lock( e );
 					monitor_cache_release( mi, e );
 				}
+				e = rs->sr_entry;
 				goto freeout;
 			}
-		} else {
+		}
+		if ( sub_nv == NULL ) {
 			monitor_cache_release( mi, e );
 		}
 
@@ -139,6 +140,7 @@ monitor_send_children(
 			rc = monitor_send_children( op, rs, sub_nv, sub_ch, sub );
 			if ( rc ) {
 freeout:
+				monitor_cache_release( mi, e );
 				if ( nonvolatile == 0 ) {
 					for ( ; e_tmp != NULL; ) {
 						mp = ( monitor_entry_t * )e_tmp->e_private;
@@ -156,6 +158,9 @@ freeout:
 				return( rc );
 			}
 		}
+		if ( sub_nv != NULL ) {
+			monitor_cache_release( mi, e );
+		}
 	}
 	
 	return LDAP_SUCCESS;
@@ -170,7 +175,7 @@ monitor_back_search( Operation *op, SlapReply *rs )
 	Entry		*e_nv = NULL, *e_ch = NULL;
 	slap_mask_t	mask;
 
-	Debug( LDAP_DEBUG_TRACE, "=> monitor_back_search\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "=> monitor_back_search\n" );
 
 
 	/* get entry with reader lock */
@@ -234,9 +239,9 @@ monitor_back_search( Operation *op, SlapReply *rs )
 	case LDAP_SCOPE_ONELEVEL:
 	case LDAP_SCOPE_SUBORDINATE:
 		monitor_find_children( op, rs, e, &e_nv, &e_ch );
-		monitor_cache_release( mi, e );
 		rc = monitor_send_children( op, rs, e_nv, e_ch,
 			op->oq_search.rs_scope == LDAP_SCOPE_SUBORDINATE );
+		monitor_cache_release( mi, e );
 		break;
 
 	case LDAP_SCOPE_SUBTREE:
@@ -245,14 +250,12 @@ monitor_back_search( Operation *op, SlapReply *rs )
 		rc = test_filter( op, e, op->oq_search.rs_filter );
 		if ( rc == LDAP_COMPARE_TRUE ) {
 			rs->sr_entry = e;
-			rs->sr_flags = REP_ENTRY_MUSTRELEASE;
 			send_search_entry( op, rs );
 			rs->sr_entry = NULL;
-		} else {
-			monitor_cache_release( mi, e );
 		}
 
 		rc = monitor_send_children( op, rs, e_nv, e_ch, 1 );
+		monitor_cache_release( mi, e );
 		break;
 
 	default:
